import json
from typing import Annotated, Any, Awaitable, Callable, List, Optional

from clowmdb.models import BucketPermission
from clowmdb.models import User as UserDB
from fastapi import APIRouter, Body, Depends, HTTPException, Query, status

from app.api.dependencies import (
    AuthorizationDependency,
    CurrentBucket,
    CurrentUser,
    DBSession,
    S3Resource,
    get_current_bucket,
    get_user_by_path_uid,
)
from app.crud.crud_bucket_permission import CRUDBucketPermission, DuplicateError
from app.schemas.bucket_permission import BucketPermissionIn as PermissionSchemaIn
from app.schemas.bucket_permission import BucketPermissionOut as PermissionSchemaOut
from app.schemas.bucket_permission import BucketPermissionParameters as PermissionParametersSchema

router = APIRouter(prefix="/permissions", tags=["BucketPermission"])
permission_authorization = AuthorizationDependency(resource="bucket_permission")
Authorization = Annotated[Callable[[str], Awaitable[Any]], Depends(permission_authorization)]


@router.get(
    "/bucket/{bucket_name}/user/{uid}",
    response_model=PermissionSchemaOut,
    summary="Get permission for bucket and user combination.",
    response_model_exclude_none=True,
)
async def get_permission_for_bucket(
    bucket: CurrentBucket,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    user: UserDB = Depends(get_user_by_path_uid),
) -> PermissionSchemaOut:
    """
    Get the bucket permissions for the specific combination of bucket and user.\n
    The owner of the bucket and the grantee of the permission can view it.\n
    Permission "bucket_permission:read" required if current user is the target or owner of the bucket permission,
    otherwise "bucket_permission:read_any" required.
    \f
    Parameters
    ----------
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    user : clowmdb.models.User
        User with the uid in the URL. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Permission for this bucket and user combination.
    """
    rbac_operation = "read" if user == current_user or current_user.uid == bucket.owner_id else "read_any"
    await authorization(rbac_operation)
    bucket_permission = await CRUDBucketPermission.get(db, bucket.name, user.uid)
    if bucket_permission:
        return PermissionSchemaOut.from_db_model(
            bucket_permission, uid=user.uid, grantee_display_name=user.display_name
        )
    raise HTTPException(
        status.HTTP_404_NOT_FOUND,
        detail=f"Permission for combination of bucket={bucket.name} and user={user.uid} doesn't exists",
    )


@router.delete(
    "/bucket/{bucket_name}/user/{uid}",
    status_code=status.HTTP_204_NO_CONTENT,
    summary="Delete a bucket permission",
)
async def delete_permission_for_bucket(
    bucket: CurrentBucket,
    db: DBSession,
    s3: S3Resource,
    current_user: CurrentUser,
    authorization: Authorization,
    user: UserDB = Depends(get_user_by_path_uid),
) -> None:
    """
    Delete the bucket permissions for the specific combination of bucket and user.\n
    The owner of the bucket and the grantee of the permission can delete it.\n
    Permission "bucket_permission:delete" required if current user is the target or owner of the bucket permission,
    otherwise "bucket_permission:delete_any" required.
    \f
    Parameters
    ----------
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    user : clowmdb.models.User
        User with the uid in the URL. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Permission for this bucket and user combination.
    """
    rbac_operation = "delete" if user == current_user or current_user.uid == bucket.owner_id else "delete_any"
    await authorization(rbac_operation)
    bucket_permission = await CRUDBucketPermission.get(db, bucket.name, user.uid)
    if bucket_permission is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail=f"Permission for combination of bucket={bucket.name} and user={user.uid} doesn't exists",
        )
    await CRUDBucketPermission.delete(db, bucket_permission)
    bucket_permission_schema = PermissionSchemaOut.from_db_model(bucket_permission, user.uid, user.display_name)
    s3_policy = s3.Bucket(bucket_permission_schema.bucket_name).Policy()
    policy = json.loads(s3_policy.policy)
    policy["Statement"] = [
        stmt for stmt in policy["Statement"] if stmt["Sid"] != bucket_permission_schema.to_hash(user.uid)
    ]
    s3_policy.put(Policy=json.dumps(policy))


@router.get(
    "/bucket/{bucket_name}",
    response_model=List[PermissionSchemaOut],
    summary="Get all permissions for a bucket.",
    response_model_exclude_none=True,
)
async def list_permissions_per_bucket(
    bucket: CurrentBucket,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    permission_type: Optional[List[BucketPermission.Permission]] = Query(
        None, description="Type of Bucket Permissions to fetch"
    ),
    permission_status: Optional[CRUDBucketPermission.PermissionStatus] = Query(
        None, description="Status of Bucket Permissions to fetch"
    ),
) -> List[PermissionSchemaOut]:
    """
    List all the bucket permissions for the given bucket.\n
    Permission "bucket_permission:read" required if current user is owner of the bucket,
    otherwise "bucket_permission:read_any" required.
    \f
    Parameters
    ----------
    permission_type : List[clowmdb.models.BucketPermission.Permission] | None, default None
        Type of Bucket Permissions to fetch. Query Parameter
    permission_status : app.crud.crud_bucket_permission.CRUDBucketPermission.PermissionStatus | None, default None
        Status of Bucket Permissions to fetch. Query Parameter.
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.

    Returns
    -------
    permissions : List[app.schemas.bucket_permission.BucketPermissionOut]
        List of all permissions for this bucket.
    """
    rbac_operation = "list_bucket" if bucket.owner_id == current_user.uid else "list_all"
    await authorization(rbac_operation)
    bucket_permissions = await CRUDBucketPermission.get_permissions_for_bucket(
        db, bucket.name, permission_type, permission_status
    )
    return [PermissionSchemaOut.from_db_model(p) for p in bucket_permissions]


@router.get(
    "/user/{uid}",
    response_model=List[PermissionSchemaOut],
    summary="Get all permissions for a user.",
    response_model_exclude_none=True,
)
async def list_permissions_per_user(
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    permission_type: Optional[List[BucketPermission.Permission]] = Query(
        None, description="Type of Bucket Permissions to fetch"
    ),
    permission_status: Optional[CRUDBucketPermission.PermissionStatus] = Query(
        None, description="Status of Bucket Permissions to fetch"
    ),
    user: UserDB = Depends(get_user_by_path_uid),
) -> List[PermissionSchemaOut]:
    """
    List all the bucket permissions for the given user.\n
    Permission "bucket_permission:read" required if current user is the target the bucket permission,
    otherwise "bucket_permission:read_any" required.
    \f
    Parameters
    ----------
    permission_type : List[clowmdb.models.BucketPermission.Permission] | None, default None
        Type of Bucket Permissions to fetch. Query Parameter
    permission_status : app.crud.crud_bucket_permission.CRUDBucketPermission.PermissionStatus | None, default None
        Status of Bucket Permissions to fetch. Query Parameter.
    user : clowmdb.models.User
        User with given uid. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.

    Returns
    -------
    permissions : List[app.schemas.bucket_permission.BucketPermissionOut]
        List of all permissions for this user.
    """
    rbac_operation = "list_user" if user == current_user else "list_all"
    await authorization(rbac_operation)
    bucket_permissions = await CRUDBucketPermission.get_permissions_for_user(
        db, user.uid, permission_type, permission_status
    )
    return [
        PermissionSchemaOut.from_db_model(p, uid=user.uid, grantee_display_name=user.display_name)
        for p in bucket_permissions
    ]


@router.post(
    "",
    response_model=PermissionSchemaOut,
    status_code=status.HTTP_201_CREATED,
    summary="Create a permission.",
    response_model_exclude_none=True,
)
async def create_permission(
    db: DBSession,
    current_user: CurrentUser,
    s3: S3Resource,
    authorization: Authorization,
    permission: PermissionSchemaIn = Body(..., description="Permission to create"),
) -> PermissionSchemaOut:
    """
    Create a permission for a bucket and user.\n
    Permission "bucket_permission:create" required.
    \f
    Parameters
    ----------
    permission : app.schemas.bucket_permission.BucketPermissionOut
        Information about the permission which should be created. HTTP Body parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Newly created permission.
    """
    await authorization("create")
    target_bucket = await get_current_bucket(permission.bucket_name, db=db)  # Check if the target bucket exists
    if target_bucket.owner_id != current_user.uid:
        raise HTTPException(status.HTTP_403_FORBIDDEN, detail="Action forbidden.")
    if target_bucket.owner_constraint is not None:
        raise HTTPException(status.HTTP_403_FORBIDDEN, detail="Initial Buckets can be target of Bucket Permissions.")
    grantee = await get_user_by_path_uid(permission.uid, db)  # Check if target user exists
    try:
        permission_db = await CRUDBucketPermission.create(db, permission)
    except ValueError:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST, detail="The owner of the bucket can't get any more permissions"
        )
    except DuplicateError:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            detail=f"Permission for combination of bucket={permission.bucket_name} and user={permission.uid} already exists",  # noqa:E501
        )
    s3_policy = s3.Bucket(permission.bucket_name).Policy()
    old_policy = s3_policy.policy
    json_policy = json.loads(old_policy)
    json_policy["Statement"] += permission.map_to_bucket_policy_statement(permission_db.user_id)
    new_policy = json.dumps(json_policy)
    s3_policy.put(Policy=new_policy)

    return PermissionSchemaOut.from_db_model(permission_db, uid=grantee.uid, grantee_display_name=grantee.display_name)


@router.put(
    "/bucket/{bucket_name}/user/{uid}",
    status_code=status.HTTP_200_OK,
    response_model=PermissionSchemaOut,
    summary="Update a bucket permission",
    response_model_exclude_none=True,
)
async def update_permission(
    bucket: CurrentBucket,
    db: DBSession,
    current_user: CurrentUser,
    s3: S3Resource,
    authorization: Authorization,
    permission_parameters: PermissionParametersSchema = Body(..., description="Permission to create"),
    user: UserDB = Depends(get_user_by_path_uid),
) -> PermissionSchemaOut:
    """
    Update a permission for a bucket and user.\n
    Permission "bucket_permission:read" required if current user is the target the bucket permission,
    otherwise "bucket_permission:update" required.
    \f
    Parameters
    ----------
    permission_parameters : app.schemas.bucket_permission.BucketPermissionOut
        Information about the permission which should be updated. HTTP Body parameter.
    user : clowmdb.models.User
        User with the uid in the URL. Dependency Injection.
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Updated permission.
    """
    await authorization("update")
    if bucket.owner_id != current_user.uid:
        raise HTTPException(status.HTTP_403_FORBIDDEN, "Action forbidden")
    bucket_permission = await CRUDBucketPermission.get(db, bucket.name, user.uid)

    if bucket_permission is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail=f"Permission for combination of bucket={bucket.name} and user={user.uid} doesn't exists",
        )
    updated_permission = await CRUDBucketPermission.update_permission(db, bucket_permission, permission_parameters)
    updated_permission_schema = PermissionSchemaOut.from_db_model(updated_permission)
    s3_policy = s3.Bucket(updated_permission_schema.bucket_name).Policy()
    policy = json.loads(s3_policy.policy)
    policy["Statement"] = [
        stmt for stmt in policy["Statement"] if stmt["Sid"] != updated_permission_schema.to_hash(user.uid)
    ]
    policy["Statement"] += updated_permission_schema.map_to_bucket_policy_statement(updated_permission.user_id)
    s3_policy.put(Policy=json.dumps(policy))
    return updated_permission_schema
