import json
from functools import reduce
from typing import TYPE_CHECKING, Annotated, Any, Awaitable, Callable, List, Optional

from botocore.exceptions import ClientError
from fastapi import APIRouter, Depends, HTTPException, Path, Query, status

from app.api.dependencies import AuthorizationDependency, CurrentBucket, CurrentUser, DBSession, S3Resource
from app.core.config import settings
from app.crud.crud_bucket import CRUDBucket
from app.crud.crud_bucket_permission import CRUDBucketPermission
from app.schemas.bucket import BucketIn as BucketInSchema
from app.schemas.bucket import BucketOut as BucketOutSchema
from app.schemas.bucket import S3ObjectMetaInformation

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import ObjectSummary
else:
    ObjectSummary = object

router = APIRouter(prefix="/buckets", tags=["Bucket"])
bucket_authorization = AuthorizationDependency(resource="bucket")
Authorization = Annotated[Callable[[str], Awaitable[Any]], Depends(bucket_authorization)]

cors_rule = {
    "CORSRules": [
        {
            "ID": "websiteaccess",
            "AllowedHeaders": [
                "amz-sdk-invocation-id",
                "amz-sdk-request",
                "authorization",
                "content-type",
                "x-amz-content-sha256",
                "x-amz-copy-source",
                "x-amz-date",
                "x-amz-user-agent",
                "content-md5",
            ],
            "AllowedMethods": ["GET", "PUT", "POST", "DELETE"],
            "AllowedOrigins": [str(settings.CLOWM_URL)[:-1]],
            "ExposeHeaders": [
                "Etag",
            ],
            "MaxAgeSeconds": 100,
        },
    ]
}


@router.get("", response_model=List[BucketOutSchema], summary="List buckets of user")
async def list_buckets(
    db: DBSession,
    s3: S3Resource,
    current_user: CurrentUser,
    authorization: Authorization,
    user: Optional[str] = Query(
        None,
        description="UID of the user for whom to fetch the buckets for. Permission 'bucket:read_any' required if current user is not the target.",  # noqa:E501
    ),
    bucket_type: CRUDBucket.BucketType = Query(
        CRUDBucket.BucketType.ALL, description="Type of the bucket to get. Ignored when `user` parameter not set"
    ),
) -> List[BucketOutSchema]:
    """
    List all the buckets in the system or of the desired user where the user has READ permissions for.\n
    Permission "bucket:read" required.
    \f
    Parameters
    ----------
    user : clowmdb.models.User
        User for which to retrieve the buckets. Dependency Injection.
    bucket_type : app.crud.crud_bucket.CRUDBucket.BucketType, default BucketType.ALL
        Type of the bucket to get. Query Parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    Returns
    -------
    buckets : List[app.schemas.bucket.BucketOut]
        All the buckets for which the user has READ permissions.
    """
    await authorization("list_all" if user is None or current_user.uid != user else "list")
    if user is None:
        buckets = await CRUDBucket.get_all(db)
    else:
        buckets = await CRUDBucket.get_for_user(db, current_user.uid, bucket_type)
    return [
        BucketOutSchema(
            **{
                "description": bucket.description,
                "name": bucket.name,
                "created_at": bucket.created_at,
                "owner": bucket.owner_id,
                "num_objects": sum(1 for obj in s3.Bucket(name=bucket.name).objects.all() if not obj.key.endswith("/")),
                "size": reduce(lambda x, y: x + y.size, list(s3.Bucket(name=bucket.name).objects.all()), 0),
                "owner_constraint": bucket.owner_constraint,
            }
        )
        for bucket in buckets
    ]


@router.post(
    "",
    response_model=BucketOutSchema,
    status_code=status.HTTP_201_CREATED,
    summary="Create a bucket for the current user",
)
async def create_bucket(
    bucket: BucketInSchema,
    current_user: CurrentUser,
    db: DBSession,
    s3: S3Resource,
    authorization: Authorization,
) -> BucketOutSchema:
    """
    Create a bucket for the current user.\n
    The name of the bucket has some constraints.
    For more information see the
    [Ceph documentation](https://docs.ceph.com/en/quincy/radosgw/s3/bucketops/#constraints)\n
    Permission "bucket:create" required.
    \f
    Parameters
    ----------
    bucket : app.schemas.bucket.BucketIn
        Information about the bucket to create. HTTP Body
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    bucket : app.schemas.bucket.BucketOut
        The newly created bucket.
    """
    await authorization("create")
    db_bucket = await CRUDBucket.create(db, bucket, current_user.uid)
    if db_bucket is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Bucket name is already taken",
        )
    s3_bucket = s3.Bucket(db_bucket.name)
    s3_bucket.create()
    # Add basic permission to the user for getting, creating and deleting objects in the bucket.
    bucket_policy = json.dumps(
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "ProxyOwnerPerm",
                    "Effect": "Allow",
                    "Principal": {"AWS": [f"arn:aws:iam:::user/{settings.BUCKET_CEPH_USERNAME}"]},
                    "Action": ["s3:GetObject"],
                    "Resource": [f"arn:aws:s3:::{db_bucket.name}/*"],
                },
                {
                    "Sid": "PseudoOwnerPerm",
                    "Effect": "Allow",
                    "Principal": {"AWS": [f"arn:aws:iam:::user/{current_user.uid}"]},
                    "Action": ["s3:GetObject", "s3:DeleteObject", "s3:PutObject", "s3:ListBucket"],
                    "Resource": [f"arn:aws:s3:::{db_bucket.name}/*", f"arn:aws:s3:::{db_bucket.name}"],
                },
            ],
        }
    )
    s3_bucket.Policy().put(Policy=bucket_policy)
    s3_bucket.Cors().put(CORSConfiguration=cors_rule)  # type: ignore[arg-type]
    return BucketOutSchema(
        **{
            "description": db_bucket.description,
            "name": db_bucket.name,
            "created_at": db_bucket.created_at,
            "owner": db_bucket.owner.uid,
            "num_objects": 0,
            "size": 0,
        }
    )


@router.get("/{bucket_name}", response_model=BucketOutSchema, summary="Get a bucket by its name")
async def get_bucket(
    bucket: CurrentBucket,
    s3: S3Resource,
    current_user: CurrentUser,
    authorization: Authorization,
    db: DBSession,
) -> BucketOutSchema:
    """
    Get a bucket by its name if the current user has READ permissions for the bucket.\n
    Permission "bucket:read" required if the current user is the owner of the bucket,
    otherwise "bucket:read_any" required.
    \f
    Parameters
    ----------
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.

    Returns
    -------
    bucket : app.schemas.bucket.BucketOut
        Bucket with the provided name.
    """
    rbac_operation = (
        "read_any"
        if not bucket.public and not await CRUDBucketPermission.check_permission(db, bucket.name, current_user.uid)
        else "read"
    )
    await authorization(rbac_operation)
    s3bucket = s3.Bucket(name=bucket.name)
    objects: List[ObjectSummary] = list(s3bucket.objects.all())
    return BucketOutSchema(
        **{
            "description": bucket.description,
            "name": bucket.name,
            "created_at": bucket.created_at,
            "owner": bucket.owner_id,
            "num_objects": sum(1 for obj in objects if not obj.key.endswith("/")),
            "size": reduce(lambda x, y: x + y.size, objects, 0),
            "owner_constraint": bucket.owner_constraint,
        }
    )


@router.delete("/{bucket_name}", status_code=status.HTTP_204_NO_CONTENT, summary="Delete a bucket")
async def delete_bucket(
    bucket: CurrentBucket,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    s3: S3Resource,
    force_delete: bool = Query(False, description="Delete even non-empty bucket"),
) -> None:
    """
    Delete a bucket by its name. Only the owner of the bucket can delete the bucket.\n
    Permission "bucket:delete" required if the current user is the owner of the bucket,
    otherwise "bucket:delete_any" required.
    \f
    Parameters
    ----------
    force_delete : bool, default False
        Flag for deleting a non-empty bucket. Query parameter.
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    """
    rbac_operation = "delete_any" if bucket.owner_id != current_user.uid else "delete"
    await authorization(rbac_operation)
    if force_delete:
        objs = [{"Key": obj.key} for obj in s3.Bucket(bucket.name).objects.all()]
        if len(objs) > 0:
            s3.Bucket(bucket.name).delete_objects(Delete={"Objects": objs})  # type: ignore
    try:
        s3.Bucket(name=bucket.name).delete()
        await CRUDBucket.delete(db, bucket)
    except ClientError:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail="Bucket not empty")


@router.get(
    "/{bucket_name}/objects",
    response_model=List[S3ObjectMetaInformation],
    tags=["Object"],
    summary="Get the metadata of the objects in the bucket",
    deprecated=True,
)
async def get_bucket_objects(
    bucket: CurrentBucket,
    s3: S3Resource,
    current_user: CurrentUser,
    db: DBSession,
    authorization: Authorization,
    file_prefix: Optional[str] = Query(None, description="Get only objects with the specified prefix."),
) -> List[S3ObjectMetaInformation]:
    """
    Get the metadata of the objects in the bucket.

    If the user has permissions for the bucket and uses the `file_prefix` parameter, it will be concatenated with the
    allowed prefix from the permission.\n
    Permission "bucket:read" required if the current user has access to the bucket,
    otherwise "bucket:read_any" required.
    \f
    Parameters
    ----------
    file_prefix : str|None, default None
        Prefix for the objects to fetch from the S3 Bucket, Query Parameter.
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    objs : List[app.schemas.bucket.S3ObjectMetaInformation]
        Meta information about all objects in the bucket.
    """
    rbac_operation = (
        "read_any" if not await CRUDBucketPermission.check_permission(db, bucket.name, current_user.uid) else "read"
    )
    await authorization(rbac_operation)
    object_prefix: str = "" if file_prefix is None else file_prefix
    permission = await CRUDBucketPermission.get(db, bucket.name, current_user.uid)
    if permission is not None and permission.file_prefix is not None:
        return [
            S3ObjectMetaInformation.from_native_s3_object(obj)
            for obj in s3.Bucket(bucket.name).objects.filter(Prefix=object_prefix + permission.file_prefix).all()
        ]
    return [
        S3ObjectMetaInformation.from_native_s3_object(obj)
        for obj in s3.Bucket(bucket.name).objects.filter(Prefix=object_prefix).all()
    ]


@router.get(
    "/{bucket_name}/objects/{object_path:path}",
    response_model=S3ObjectMetaInformation,
    tags=["Object"],
    summary="Get the metadata about a specific object",
    deprecated=True,
)
async def get_bucket_object(
    bucket: CurrentBucket,
    s3: S3Resource,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    object_path: str = Path(..., decsription="Name of the object", examples=["test.txt", "pseudo/sub/folder/test.txt"]),
) -> S3ObjectMetaInformation:
    """
    Get the metadata of a specific object in a bucket.\n
    Permission "bucket:read" required if the current user has access to the bucket,
    otherwise "bucket:read_any" required.
    \f
    Parameters
    ----------
    bucket : clowmdb.models.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    object_path : str
        Key of a specific object in the bucket. URL Path Parameter.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    objs : app.schemas.bucket.S3ObjectMetaInformation
        Meta information about the specific object in the bucket.
    """
    rbac_operation = (
        "read" if await CRUDBucketPermission.check_permission(db, bucket.name, current_user.uid) else "read_any"
    )
    await authorization(rbac_operation)
    permission = await CRUDBucketPermission.get(db, bucket.name, current_user.uid)
    try:
        if permission is None or permission.file_prefix is None or object_path.startswith(permission.file_prefix):
            obj = s3.ObjectSummary(bucket_name=bucket.name, key=object_path)
            return S3ObjectMetaInformation.from_native_s3_object(obj)
        raise HTTPException(status.HTTP_403_FORBIDDEN, detail="No rights for this object.")
    except ClientError:
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Object not found")
