from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import AnyHttpUrl, AnyUrl, Field, computed_field
from pydantic_settings import BaseSettings, SettingsConfigDict


def _assemble_db_uri(values: Dict[str, Any], async_flag: bool = True) -> Any:
    return AnyUrl.build(
        scheme=f"mysql+{'aiomysql' if async_flag else 'pymysql'}",
        password=values.get("DB_PASSWORD"),
        username=values.get("DB_USER"),
        port=values.get("DB_PORT"),
        host=values.get("DB_HOST"),  # type: ignore[arg-type]
        path=f"{values.get('DB_DATABASE') or ''}",
    )


def _load_public_key(pub_key_val: Optional[str], pub_key_file: Optional[Path]) -> str:
    pub_key = ""
    if pub_key_val is not None:
        pub_key = pub_key_val
    if pub_key_file is not None:
        with open(pub_key_file) as f:
            pub_key = f.read()
    if len(pub_key) == 0:
        raise ValueError("PUBLIC_KEY_VALUE or PUBLIC_KEY_FILE must be set")
    return pub_key


class Settings(BaseSettings):
    CLOWM_URL: AnyHttpUrl = Field(
        AnyHttpUrl("http://localhost:8080"),
        description="Base HTTP URL where the CloWM service is reachable.",
        examples=["http://localhost:8080"],
    )
    API_PREFIX: str = Field("/api", description="Path Prefix for all API endpoints.")

    public_key_value: Optional[str] = Field(
        None, description="Public RSA Key in PEM format to sign the JWTs.", validation_alias="PUBLIC_KEY_VALUE"
    )
    public_key_file: Optional[Path] = Field(
        None, description="Path to Public RSA Key in PEM format to sign the JWTs.", validation_alias="PUBLIC_KEY_FILE"
    )

    @computed_field  # type: ignore[misc]
    @property
    def PUBLIC_KEY(self) -> str:
        return _load_public_key(self.public_key_value, self.public_key_file)

    # BACKEND_CORS_ORIGINS is a JSON-formatted list of origins
    BACKEND_CORS_ORIGINS: List[AnyHttpUrl] = Field([], description="List of all valid CORS origins")

    DB_HOST: str = Field(..., description="Host of the database.")
    DB_USER: str = Field(..., description="Username in the database.")
    DB_PASSWORD: str = Field(..., description="Password for the database user.")
    DB_DATABASE: str = Field(..., description="Name of the database.")
    DB_PORT: int = Field(3306, description="Port of the database.")
    SQLALCHEMY_VERBOSE_LOGGER: bool = Field(False, description="Flag whether to print the SQL Queries in the logs.")

    @computed_field  # type: ignore[misc]
    @property
    def SQLALCHEMY_DATABASE_ASYNC_URI(self) -> AnyUrl:
        return _assemble_db_uri(
            {
                "DB_HOST": self.DB_HOST,
                "DB_USER": self.DB_USER,
                "DB_PASSWORD": self.DB_PASSWORD,
                "DB_DATABASE": self.DB_DATABASE,
                "DB_PORT": self.DB_PORT,
            },
            async_flag=True,
        )

    @computed_field  # type: ignore[misc]
    @property
    def SQLALCHEMY_DATABASE_NORMAL_URI(self) -> AnyUrl:
        return _assemble_db_uri(
            {
                "DB_HOST": self.DB_HOST,
                "DB_USER": self.DB_USER,
                "DB_PASSWORD": self.DB_PASSWORD,
                "DB_DATABASE": self.DB_DATABASE,
                "DB_PORT": self.DB_PORT,
            },
            async_flag=False,
        )

    OBJECT_GATEWAY_URI: AnyHttpUrl = Field(..., description="URI of the Ceph Object Gateway.")
    USER_CEPH_ACCESS_KEY: str = Field(
        ..., description="Access key for the Ceph Object Gateway with 'user:read,user:write privileges'."
    )
    USER_CEPH_SECRET_KEY: str = Field(
        ..., description="Secret key for the Ceph Object Gateway with 'user:read,user:write privileges'."
    )
    BUCKET_CEPH_ACCESS_KEY: str = Field(
        ..., description="Access key for the Ceph Object Gateway with unlimited buckets."
    )
    BUCKET_CEPH_SECRET_KEY: str = Field(
        ..., description="Secret key for the Ceph Object Gateway with unlimited buckets."
    )
    BUCKET_CEPH_USERNAME: str = Field(
        ..., description="ID of the user in ceph who owns all the buckets. Owner of 'BUCKET_CEPH_ACCESS_KEY'"
    )
    OPA_URI: AnyHttpUrl = Field(..., description="URI of the OPA Service")
    OPA_POLICY_PATH: str = Field("/clowm/authz/allow", description="Path to the OPA Policy for Authorization")
    model_config = SettingsConfigDict(case_sensitive=True, env_file=".env", secrets_dir="/run/secrets")


settings = Settings()
