from typing import Dict

from authlib.jose import JsonWebToken
from fastapi import HTTPException, status
from httpx import AsyncClient

from app.core.config import settings
from app.schemas.security import AuthzRequest, AuthzResponse

ISSUER = "clowm"
ALGORITHM = "RS256"
jwt = JsonWebToken([ALGORITHM])


def decode_token(token: str) -> Dict[str, str]:  # pragma: no cover
    """
    Decode and verify a JWT token.

    Parameters
    ----------
    token : str
        The JWT to decode.

    Returns
    -------
    decoded_token : Dict[str, str]
        Payload of the decoded token.
    """
    claims = jwt.decode(
        s=token,
        key=settings.PUBLIC_KEY,
        claims_options={
            "iss": {"essential": True},
            "sub": {"essential": True},
            "exp": {"essential": True},
        },
    )
    claims.validate()
    return claims


async def request_authorization(request_params: AuthzRequest, client: AsyncClient) -> AuthzResponse:
    """
    Send a request to OPA for a policy decision. Raise an HTTPException with status 403 if authorization is denied.

    Parameters
    ----------
    request_params : app.schemas.security.AuthRequest
        Input parameters for the authorization request.
    client : httpx.AsyncClient
        An async http client with an open connection. This function doesn't close the connection afterwards.

    Returns
    -------
    response : app.schemas.security.AuthzResponse
        Response by the Auth service about the authorization request
    """
    response = await client.post(
        f"{settings.OPA_URI}v1/data{settings.OPA_POLICY_PATH}", json={"input": request_params.model_dump()}
    )

    parsed_response = AuthzResponse(**response.json())
    if not parsed_response.result:  # pragma: no cover
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN, detail=f"Action forbidden. Decision ID {parsed_response.decision_id}"
        )
    return parsed_response
