from enum import Enum, unique
from typing import Optional, Sequence

from clowmdb.models import Bucket
from clowmdb.models import BucketPermission as BucketPermissionDB
from sqlalchemy import func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession

from app.schemas.bucket import BucketIn as BucketInSchema


class CRUDBucket:
    @unique
    class BucketType(str, Enum):
        """
        Enumeration for the type of buckets to fetch from the DB

        OWN: Only fetch buckets that the user owns
        PERMISSION: Only fetch foreign buckets that the user has access to
        ALL: Fetch all buckets that the user has access to
        """

        OWN: str = "OWN"
        ALL: str = "ALL"
        PERMISSION: str = "PERMISSION"

    @staticmethod
    async def get(db: AsyncSession, bucket_name: str) -> Optional[Bucket]:
        """
        Get a bucket by its name.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of the Bucket to get from database.

        Returns
        -------
        bucket : Bucket | None
            Returns the bucket if it exists, None otherwise.
        """
        stmt = select(Bucket).where(Bucket.name == bucket_name)
        row = await db.execute(stmt)
        return row.scalar()

    @staticmethod
    async def get_all(db: AsyncSession) -> Sequence[Bucket]:
        stmt = select(Bucket)
        buckets = (await db.execute(stmt)).scalars().all()
        return buckets

    @staticmethod
    async def get_for_user(db: AsyncSession, uid: str, bucket_type: BucketType = BucketType.ALL) -> Sequence[Bucket]:
        """
        Get all buckets for a user. Depending on the `bucket_type`, the user is either owner of the bucket or has
        permission for the bucket

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        uid : str
            UID of a user.
        bucket_type : BucketType, default BucketType.ALL


        Returns
        -------
        buckets : List[clowmdb.models.Bucket]
            A list of all buckets where the given user has READ permissions for.

        Notes
        -----
        SQL Query own buckets:
        ```
        SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id
        FROM bucket
        WHERE bucket.owner_id = %s
        ```

        SQL Query all buckets that the user has access to:
        ```
        SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id
        FROM bucket
        WHERE bucket.owner_id = %s OR (EXISTS
            (SELECT 1 FROM bucketpermission
            WHERE bucket.name = bucketpermission.bucket_name AND bucketpermission.user_id = %s
                AND(UNIX_TIMESTAMP() >= bucketpermission.`from` 0 OR bucketpermission.`from` IS NULL)
                AND(UNIX_TIMESTAMP() <= bucketpermission.`to` >= 0 OR bucketpermission.`to` IS NULL)))
        ```

        SQL Query only foreign buckets where user has permission to
        ```
        SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id
        FROM bucket
        WHERE (EXISTS
            (SELECT 1 FROM bucketpermission
            WHERE bucket.name = bucketpermission.bucket_name AND bucketpermission.user_id = %s
                AND(UNIX_TIMESTAMP() >=  bucketpermission.`from` <= 0 OR bucketpermission.`from` IS NULL)
                AND(UNIX_TIMESTAMP() <=  bucketpermission.`to` >= 0 OR bucketpermission.`to` IS NULL)))
        ```
        """
        stmt = select(Bucket)
        if bucket_type == CRUDBucket.BucketType.ALL:
            stmt = stmt.where(
                or_(
                    Bucket.owner_id == uid,
                    Bucket.permissions.any(BucketPermissionDB.user_id == uid)
                    .where(
                        or_(
                            func.UNIX_TIMESTAMP() >= BucketPermissionDB.from_,
                            BucketPermissionDB.from_ == None,  # noqa:E711
                        )
                    )
                    .where(
                        or_(
                            func.UNIX_TIMESTAMP() <= BucketPermissionDB.to,
                            BucketPermissionDB.to == None,  # noqa:E711
                        )
                    ),
                )
            )
        elif bucket_type == CRUDBucket.BucketType.OWN:
            stmt = stmt.where(Bucket.owner_id == uid)
        else:
            stmt = stmt.where(
                Bucket.permissions.any(BucketPermissionDB.user_id == uid)
                .where(
                    or_(
                        func.UNIX_TIMESTAMP() >= BucketPermissionDB.from_,
                        BucketPermissionDB.from_ == None,  # noqa:E711
                    )
                )
                .where(
                    or_(
                        func.UNIX_TIMESTAMP() <= BucketPermissionDB.to,
                        BucketPermissionDB.to == None,  # noqa:E711
                    )
                ),
            )
        buckets = (await db.execute(stmt)).scalars().all()
        return buckets

    @staticmethod
    async def create(db: AsyncSession, bucket_in: BucketInSchema, uid: str) -> Optional[Bucket]:
        """
        Create a bucket for a given user.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket_in : app.schemas.bucket.BucketIn
            All relevant information for a new bucket.
        uid : str
            UID of the owner for the new bucket.

        Returns
        -------
        bucket : clowmdb.models.Bucket | None
            Returns the created bucket. If None then there was a problem, e.g. the name of the bucket is already taken.
        """
        bucket = Bucket(**bucket_in.model_dump(), owner_id=uid)
        if await CRUDBucket.get(db, bucket.name) is None:
            db.add(bucket)
            await db.commit()
            await db.refresh(bucket)
            return bucket
        return None

    @staticmethod
    async def delete(db: AsyncSession, bucket: Bucket) -> None:
        """
        Delete a specific bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket : clowmdb.models.Bucket
            The bucket to delete.
        """
        await db.delete(bucket)
        await db.commit()
