from enum import Enum, unique
from typing import List, Optional, Sequence

from clowmdb.models import BucketPermission as BucketPermissionDB
from sqlalchemy import and_, func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import Select as SQLSelect

from app.crud.crud_bucket import CRUDBucket
from app.crud.crud_user import CRUDUser
from app.schemas.bucket_permission import BucketPermissionIn as BucketPermissionSchema
from app.schemas.bucket_permission import BucketPermissionParameters as BucketPermissionParametersSchema


class CRUDBucketPermission:
    @unique
    class PermissionStatus(str, Enum):
        """
        Status of a bucket permission. Can be either `ACTIVE` or `INACTIVE`. A permission can only get `INACTIVE` if the
        permission itself has a time limit and the current time is not in the timespan.
        """

        ACTIVE: str = "ACTIVE"
        INACTIVE: str = "INACTIVE"

    @staticmethod
    async def get(db: AsyncSession, bucket_name: str, user_id: str) -> Optional[BucketPermissionDB]:
        stmt = select(BucketPermissionDB).where(
            and_(BucketPermissionDB.user_id == user_id, BucketPermissionDB.bucket_name == bucket_name)
        )
        row = await db.execute(stmt)
        return row.scalar()

    @staticmethod
    async def get_permissions_for_bucket(
        db: AsyncSession,
        bucket_name: str,
        permission_types: Optional[List[BucketPermissionDB.Permission]] = None,
        permission_status: Optional[PermissionStatus] = None,
    ) -> Sequence[BucketPermissionDB]:
        """
        Get the permissions for the given bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of the bucket which to query.
        permission_types : List[clowmdb.models.BucketPermission.Permission] | None, default None
            Type of Bucket Permissions to fetch.
        permission_status : app.crud.crud_bucket_permission.CRUDBucketPermission.PermissionStatus | None, default None
            Status of Bucket Permissions to fetch.

        Returns
        -------
        buckets : List[BucketPermission]
            Returns the permissions for the given bucket.
        """
        stmt = (
            select(BucketPermissionDB)
            .options(joinedload(BucketPermissionDB.grantee))
            .where(BucketPermissionDB.bucket_name == bucket_name)
        )
        if permission_types is not None:
            stmt = CRUDBucketPermission._filter_permission_types(stmt, permission_types)
        if permission_status is not None:
            stmt = CRUDBucketPermission._filter_permission_status(stmt, permission_status)
        row = await db.execute(stmt)
        return row.scalars().all()

    @staticmethod
    async def get_permissions_for_user(
        db: AsyncSession,
        user_id: str,
        permission_types: Optional[List[BucketPermissionDB.Permission]] = None,
        permission_status: Optional[PermissionStatus] = None,
    ) -> Sequence[BucketPermissionDB]:
        """
        Get the permissions for the given user.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        user_id : str
            UID of the user which to query.
        permission_types : List[clowmdb.models.BucketPermission.Permission] | None, default None
            Type of Bucket Permissions to fetch.
        permission_status : app.crud.crud_bucket_permission.CRUDBucketPermission.PermissionStatus | None, default None
            Status of Bucket Permissions to fetch.

        Returns
        -------
        buckets : List[BucketPermission]
            Returns the permissions for the given user.
        """
        stmt = select(BucketPermissionDB).where(BucketPermissionDB.user_id == user_id)
        if permission_types is not None:
            stmt = CRUDBucketPermission._filter_permission_types(stmt, permission_types)
        if permission_status is not None:
            stmt = CRUDBucketPermission._filter_permission_status(stmt, permission_status)
        row = await db.execute(stmt)
        return row.scalars().all()

    @staticmethod
    async def check_permission(db: AsyncSession, bucket_name: str, uid: str) -> bool:
        """
        Check if the provided user has any permission to the provided bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket_name : str
            Name of the bucket for which to perform the check.
        uid : str
            UID of the user for which to perform the check.

        Returns
        -------
        permission_check : bool
            Return True if the user has any permission on the bucket, False otherwise.
        """
        buckets = await CRUDBucket.get_for_user(db, uid, bucket_type=CRUDBucket.BucketType.ALL)
        return bucket_name in map(lambda x: x.name, buckets)

    @staticmethod
    async def create(db: AsyncSession, permission: BucketPermissionSchema) -> BucketPermissionDB:
        """
        Create a permission in the database and raise Exceptions if there are problems.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        permission : app.schemas.bucket_permission.BucketPermissionOut
            The permission to create.
        Returns
        -------
        permission : clowmdb.models.BucketPermission
            Newly created permission model from the db.
        """
        # Check if user exists
        user = await CRUDUser.get(db, uid=permission.uid)
        if user is None:
            raise KeyError(
                f"Unknown user with uid {permission.uid}",
            )
        # Check that grantee is not the owner of the bucket
        bucket = await CRUDBucket.get(db, permission.bucket_name)
        if bucket is None or bucket.owner_id == user.uid:
            raise ValueError(f"User {permission.uid} is the owner of the bucket {permission.bucket_name}")
        # Check if combination of user and bucket already exists
        duplicate_check_stmt = select(BucketPermissionDB).where(
            and_(BucketPermissionDB.user_id == user.uid, BucketPermissionDB.bucket_name == permission.bucket_name)
        )
        previous_permission = (await db.execute(duplicate_check_stmt)).scalar()
        if previous_permission is not None:
            raise DuplicateError(
                f"bucket permission for combination {permission.bucket_name} {permission.uid} already exists."
            )
        # Add permission to db
        permission_db = BucketPermissionDB(
            user_id=user.uid,
            bucket_name=permission.bucket_name,
            from_=permission.from_timestamp,
            to=permission.to_timestamp,
            file_prefix=permission.file_prefix,
            permissions=permission.permission,
        )
        db.add(permission_db)
        await db.commit()
        await db.refresh(permission_db)
        return permission_db

    @staticmethod
    async def delete(db: AsyncSession, permission: BucketPermissionDB) -> None:
        """
        Delete a permission in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        permission : app.schemas.bucket_permission.BucketPermissionOut
            The permission to create.
        Returns
        -------
        permission : clowmdb.models.BucketPermission
            Newly created permission model from the db.
        """
        await db.delete(permission)
        await db.commit()

    @staticmethod
    async def update_permission(
        db: AsyncSession, permission: BucketPermissionDB, new_params: BucketPermissionParametersSchema
    ) -> BucketPermissionDB:
        """
        Update a permission in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        permission : app.schemas.bucket_permission.BucketPermissionOut
            The permission to update.
        new_params : app.schemas.bucket_permission.BucketPermissionParameters
            The parameters which should be updated.

        Returns
        -------
        permission : clowmdb.models.BucketPermission
            Updated permission model from the db.
        """
        permission.from_ = new_params.from_timestamp
        permission.to = new_params.to_timestamp
        permission.file_prefix = new_params.file_prefix
        permission.permissions = new_params.permission
        await db.commit()
        await db.refresh(permission)
        return permission

    @staticmethod
    def _filter_permission_types(stmt: SQLSelect, permission_types: List[BucketPermissionDB.Permission]) -> SQLSelect:
        """
        Add a where clauses to the SQL Statement where the type of permission is filtered.

        Parameters
        ----------
        stmt : sqlalchemy.sql.Select
            Declarative Select statement from SQLAlchemy.
        permission_types : List[clowmdb.models.BucketPermission.Permission]
            Type of Bucket Permissions to filter for.

        Returns
        -------
        stmt : sqlalchemy.sql.Select
            Declarative Select statement with added where clause.
        """
        return stmt.where(or_(*[BucketPermissionDB.permissions == p_type for p_type in permission_types]))

    @staticmethod
    def _filter_permission_status(stmt: SQLSelect, permission_status: PermissionStatus) -> SQLSelect:
        """
        Add a where clauses to the SQL Statement where the status of permission is filtered based on the current time.

        Parameters
        ----------
        stmt : sqlalchemy.sql.Select
            Declarative Select statement from SQLAlchemy
        permission_status : PermissionStatus
            Status of Bucket Permissions to filter for.

        Returns
        -------
        stmt : sqlalchemy.sql.Select
            Declarative Select statement with added where clause.
        """
        if permission_status == CRUDBucketPermission.PermissionStatus.ACTIVE:
            return stmt.where(
                or_(
                    func.UNIX_TIMESTAMP() >= BucketPermissionDB.from_,
                    BucketPermissionDB.from_ == None,  # noqa:E711
                )
            ).where(
                or_(
                    func.UNIX_TIMESTAMP() <= BucketPermissionDB.to,
                    BucketPermissionDB.to == None,  # noqa:E711
                )
            )
        else:
            return stmt.where(
                or_(
                    and_(
                        func.UNIX_TIMESTAMP() <= BucketPermissionDB.from_,
                        BucketPermissionDB.from_ != None,  # noqa:E711
                    ),
                    and_(
                        func.UNIX_TIMESTAMP() >= BucketPermissionDB.to,
                        BucketPermissionDB.to != None,  # noqa:E711
                    ),
                )
            )


class DuplicateError(Exception):
    pass
