import hashlib
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from clowmdb.models import BucketPermission as BucketPermissionDB
from pydantic import BaseModel, Field


class BucketPermissionParameters(BaseModel):
    """
    Schema for the parameters of a bucket permission.
    """

    from_timestamp: Optional[int] = Field(
        None,
        description="Start date of permission as UNIX timestamp",
        examples=[1640991600],  # 01.01.2022 00:00
    )
    to_timestamp: Optional[int] = Field(
        None,
        description="End date of permission as UNIX timestamp",
        examples=[1640991600],  # 01.01.2022 00:00
    )
    file_prefix: Optional[str] = Field(None, description="Prefix of subfolder", examples=["pseudo/sub/folder/"])
    permission: Union[BucketPermissionDB.Permission, str] = Field(
        BucketPermissionDB.Permission.READ, description="Permission", examples=[BucketPermissionDB.Permission.READ]
    )


class BucketPermissionIn(BucketPermissionParameters):
    uid: str = Field(..., description="UID of the grantee", examples=["28c5353b8bb34984a8bd4169ba94c606"])
    bucket_name: str = Field(..., description="Name of Bucket", examples=["test-bucket"])

    def to_hash(self, user_id: str) -> str:
        """
        Combine the bucket name and user id and produce the MD5 hash of it.

        Parameters
        ----------
        user_id : str
            The unique and unchanging user id

        Returns
        -------
        hash : str
            The resulting MD5 hash.
        """
        str_for_id_hash = self.bucket_name + user_id
        return hashlib.md5(str_for_id_hash.encode("utf-8")).hexdigest()

    def map_to_bucket_policy_statement(self, user_id: str) -> List[Dict[str, Any]]:
        """
        Create a bucket policy statement from the schema and the user_id.\n
        The Sid is unique for every bucket and user combination.

        Parameters
        ----------
        user_id : str
            The unique and unchanging user id belonging to this permission.

        Returns
        -------
        statements : List[Dict[str, Any]]
            Bucket and object permission statements.
        """
        obj_policy: Dict[str, Any] = {
            "Sid": self.to_hash(user_id),
            "Effect": "Allow",
            "Principal": {"AWS": f"arn:aws:iam:::user/{self.uid}"},
            "Resource": f"arn:aws:s3:::{self.bucket_name}/{'' if self.file_prefix is None else self.file_prefix}*",
            "Action": [],
            "Condition": {},
        }
        bucket_policy: Dict[str, Any] = {
            "Sid": self.to_hash(user_id),
            "Effect": "Allow",
            "Principal": {"AWS": f"arn:aws:iam:::user/{self.uid}"},
            "Resource": f"arn:aws:s3:::{self.bucket_name}",
            "Action": [],
            "Condition": {},
        }
        bucket_policy["Action"] += ["s3:ListBucket"]
        if (
            self.permission == BucketPermissionDB.Permission.READ
            or self.permission == BucketPermissionDB.Permission.READWRITE
        ):
            obj_policy["Action"] += ["s3:GetObject"]
        if (
            self.permission == BucketPermissionDB.Permission.WRITE
            or self.permission == BucketPermissionDB.Permission.READWRITE
        ):
            obj_policy["Action"] += ["s3:DeleteObject", "s3:PutObject"]
            bucket_policy["Action"] += ["s3:DeleteObject"]
        if self.to_timestamp is not None:
            print(self.to_timestamp)
            obj_policy["Condition"]["DateLessThan"] = {
                "aws:CurrentTime": datetime.fromtimestamp(self.to_timestamp).strftime("%Y-%m-%dT%H:%M:%SZ")
            }
            bucket_policy["Condition"]["DateLessThan"] = obj_policy["Condition"]["DateLessThan"]
        if self.from_timestamp is not None:
            print(self.from_timestamp)
            obj_policy["Condition"]["DateGreaterThan"] = {
                "aws:CurrentTime": datetime.fromtimestamp(self.from_timestamp).strftime("%Y-%m-%dT%H:%M:%SZ")
            }
            bucket_policy["Condition"]["DateGreaterThan"] = obj_policy["Condition"]["DateGreaterThan"]
        if self.file_prefix is not None:
            bucket_policy["Condition"]["StringLike"] = {"s3:prefix": self.file_prefix + "*"}
        if len(bucket_policy["Condition"]) == 0:
            del bucket_policy["Condition"]
        if len(obj_policy["Condition"]) == 0:
            del obj_policy["Condition"]
        return [obj_policy, bucket_policy]


class BucketPermissionOut(BucketPermissionIn):
    """
    Schema for the bucket permissions.
    """

    grantee_display_name: str = Field(..., description="Display Name of the grantee", examples=["Bilbo Baggins"])

    @staticmethod
    def from_db_model(
        permission: BucketPermissionDB, uid: Optional[str] = None, grantee_display_name: Optional[str] = None
    ) -> "BucketPermissionOut":
        """
        Create a bucket permission schema from the database model.

        Parameters
        ----------
        permission : clowmdb.models.BucketPermission
            DB model for the permission.
        uid : str | None, default None
            Sets the uid in the schema. If None it will be taken from the database model.
        grantee_display_name: str | None, default None
            Sets the display name of the grantee in the schema. If None it will be taken from the database model.

        Returns
        -------
        permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Schema populated with the values from the database model.
        """
        return BucketPermissionOut(
            uid=uid if uid else permission.grantee.uid,
            grantee_display_name=grantee_display_name if grantee_display_name else permission.grantee.display_name,
            bucket_name=permission.bucket_name,
            from_timestamp=permission.from_,
            to_timestamp=permission.to,
            file_prefix=permission.file_prefix,
            permission=permission.permissions,
        )
