import pytest
from clowmdb.models import Bucket, BucketPermission
from fastapi import status
from httpx import AsyncClient
from sqlalchemy.ext.asyncio import AsyncSession

from app.crud.crud_bucket import CRUDBucket
from app.schemas.bucket import BucketIn
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.bucket import add_permission_for_bucket
from app.tests.utils.user import UserWithAuthHeader
from app.tests.utils.utils import random_lower_string


class _TestBucketRoutes:
    base_path = "/buckets"


class TestBucketRoutesGet(_TestBucketRoutes):
    @pytest.mark.asyncio
    async def test_get_all_buckets(
        self, client: AsyncClient, random_bucket: Bucket, random_second_user: UserWithAuthHeader
    ) -> None:
        """
        Test for getting all buckets with "list_all" operation.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        response = await client.get(f"{self.base_path}", headers=random_second_user.auth_headers)
        assert response.status_code == status.HTTP_200_OK

        buckets = response.json()

        assert len(buckets) == 1
        bucket = buckets[0]

        assert bucket["name"].split(":")[-1] == random_bucket.name
        assert bucket["owner"] == random_bucket.owner_id

    @pytest.mark.asyncio
    async def test_get_own_buckets(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for getting the buckets where the user is the owner.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        response = await client.get(
            f"{self.base_path}", params={"user": random_bucket.owner_id}, headers=random_user.auth_headers
        )
        assert response.status_code == status.HTTP_200_OK

        buckets = response.json()

        assert len(buckets) == 1
        bucket = buckets[0]

        assert bucket["name"].split(":")[-1] == random_bucket.name
        assert bucket["owner"] == random_bucket.owner_id

    @pytest.mark.asyncio
    async def test_get_bucket_by_name(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for getting a bucket by its name where the user is the owner of the bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        response = await client.get(f"{self.base_path}/{random_bucket.name}", headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_200_OK

        bucket = response.json()

        assert bucket["name"].split(":")[-1] == random_bucket.name
        assert bucket["owner"] == random_bucket.owner.uid

    @pytest.mark.asyncio
    async def test_get_unknown_bucket(self, client: AsyncClient, random_user: UserWithAuthHeader) -> None:
        """
        Test for getting an unknown bucket by its name.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        response = await client.get(f"{self.base_path}/impossible_bucket_name", headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_get_foreign_bucket(
        self, client: AsyncClient, random_bucket: Bucket, random_second_user: UserWithAuthHeader
    ) -> None:
        """
        Test for getting a foreign bucket with permission by its name.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random user which is not the owner of the bucket. pytest fixture.
        """
        response = await client.get(f"{self.base_path}/{random_bucket.name}", headers=random_second_user.auth_headers)
        assert response.status_code == status.HTTP_403_FORBIDDEN


class TestBucketRoutesCreate(_TestBucketRoutes):
    @pytest.mark.asyncio
    async def test_create_bucket(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for creating a bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        bucket_info = BucketIn(name=random_lower_string(), description=random_lower_string(127))
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=bucket_info.model_dump())

        assert response.status_code == status.HTTP_201_CREATED
        bucket = response.json()
        assert bucket
        assert bucket["name"].split(":")[-1] == bucket_info.name
        assert bucket["owner"] == random_user.user.uid

        dbBucket = await CRUDBucket.get(db, bucket_info.name)
        assert dbBucket
        assert dbBucket.name == bucket_info.name
        assert dbBucket.owner_id == random_user.user.uid

        await CRUDBucket.delete(db, dbBucket)

    @pytest.mark.asyncio
    async def test_create_duplicated_bucket(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for creating a bucket where the name is already taken.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        bucket_info = BucketIn(name=random_bucket.name, description=random_lower_string(127))
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=bucket_info.model_dump())

        assert response.status_code == status.HTTP_400_BAD_REQUEST


class TestBucketRoutesDelete(_TestBucketRoutes):
    @pytest.mark.asyncio
    async def test_delete_empty_bucket(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for deleting an empty bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        """
        response = await client.delete(
            f"{self.base_path}/{random_bucket.name}", headers=random_user.auth_headers, params={"force_delete": False}
        )

        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.asyncio
    async def test_delete_foreign_bucket_with_permission(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for deleting a foreign bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random user which is not the owner of the bucket. pytest fixture.
        """
        bucket = Bucket(
            name=random_lower_string(),
            description=random_lower_string(127),
            owner_id=random_second_user.user.uid,
        )
        db.add(bucket)
        await db.commit()
        await add_permission_for_bucket(
            db, bucket.name, random_user.user.uid, permission=BucketPermission.Permission.READWRITE
        )

        response = await client.delete(
            f"{self.base_path}/{bucket.name}", headers=random_user.auth_headers, params={"force_delete": False}
        )

        assert response.status_code == status.HTTP_403_FORBIDDEN

        await db.delete(bucket)

    @pytest.mark.asyncio
    async def test_delete_non_empty_bucket(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for deleting a non-empty bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects. pytest fixture.
        """
        mock_s3_service.create_object_in_bucket(bucket_name=random_bucket.name, key=random_lower_string())
        response = await client.delete(
            f"{self.base_path}/{random_bucket.name}", headers=random_user.auth_headers, params={"force_delete": False}
        )

        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_force_delete_non_empty_bucket(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for force deleting a non-empty bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing. pytest fixture.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects. pytest fixture.
        """
        mock_s3_service.create_object_in_bucket(bucket_name=random_bucket.name, key=random_lower_string())
        response = await client.delete(
            f"{self.base_path}/{random_bucket.name}", headers=random_user.auth_headers, params={"force_delete": True}
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT
