import asyncio
import json
from functools import partial
from secrets import token_urlsafe
from typing import AsyncGenerator, Callable, Dict, Generator
from uuid import uuid4

import httpx
import pytest
import pytest_asyncio
from clowmdb.db.session import get_async_session
from clowmdb.models import Bucket
from clowmdb.models import BucketPermission as BucketPermissionDB
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import get_decode_jwt_function, get_httpx_client, get_rgw_admin, get_s3_resource
from app.core.config import settings
from app.main import app
from app.schemas.bucket_permission import BucketPermissionOut as BucketPermissionSchema
from app.schemas.security import AuthzResponse
from app.tests.mocks.mock_rgw_admin import MockRGWAdmin
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.bucket import create_random_bucket
from app.tests.utils.user import UserWithAuthHeader, create_random_user, decode_mock_token, get_authorization_headers
from app.tests.utils.utils import request_admin_permission

jwt_secret = token_urlsafe(32)


@pytest.fixture(scope="session")
def event_loop() -> Generator:
    """
    Creates an instance of the default event loop for the test session.
    """
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def mock_rgw_admin() -> MockRGWAdmin:
    """
    Fixture for creating a mock object for the rgwadmin package.
    """
    return MockRGWAdmin()


@pytest.fixture(scope="session")
def mock_s3_service() -> MockS3ServiceResource:
    """
    Fixture for creating a mock object for the rgwadmin package.
    """
    return MockS3ServiceResource()


@pytest_asyncio.fixture(scope="module")
async def client(mock_rgw_admin: MockRGWAdmin, mock_s3_service: MockS3ServiceResource) -> AsyncGenerator:
    """
    Fixture for creating a TestClient and perform HTTP Request on it.
    Overrides the dependency for the RGW admin operations.
    """

    def get_mock_rgw() -> MockRGWAdmin:
        return mock_rgw_admin

    def get_mock_s3() -> MockS3ServiceResource:
        return mock_s3_service

    def get_decode_token_function() -> Callable[[str], Dict[str, str]]:
        # Override the decode_jwt function with mock function for tests and inject random shared secret
        return partial(decode_mock_token, secret=jwt_secret)

    async def get_mock_httpx_client() -> AsyncGenerator[httpx.AsyncClient, None]:
        def mock_request_handler(request: httpx.Request) -> httpx.Response:
            response_body = {}
            if str(request.url).startswith(str(settings.OPA_URI)):
                response_body = AuthzResponse(
                    result=not request_admin_permission(request), decision_id=str(uuid4())
                ).model_dump()
            return httpx.Response(200, json=response_body)

        async with httpx.AsyncClient(transport=httpx.MockTransport(mock_request_handler)) as http_client:
            yield http_client

    app.dependency_overrides[get_httpx_client] = get_mock_httpx_client
    app.dependency_overrides[get_rgw_admin] = get_mock_rgw
    app.dependency_overrides[get_s3_resource] = get_mock_s3
    app.dependency_overrides[get_decode_jwt_function] = get_decode_token_function
    async with httpx.AsyncClient(app=app, base_url="http://localhost") as ac:
        yield ac
    app.dependency_overrides = {}


@pytest_asyncio.fixture(scope="module")
async def db() -> AsyncGenerator[AsyncSession, None]:
    """
    Fixture for creating a database session to connect to.
    """
    async with get_async_session(
        url=str(settings.SQLALCHEMY_DATABASE_ASYNC_URI), verbose=settings.SQLALCHEMY_VERBOSE_LOGGER
    ) as dbSession:
        yield dbSession


@pytest_asyncio.fixture(scope="module")
async def random_user(db: AsyncSession, mock_rgw_admin: MockRGWAdmin) -> AsyncGenerator[UserWithAuthHeader, None]:
    """
    Create a random user and deletes him afterwards.
    """
    user = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user.uid)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    mock_rgw_admin.delete_user(uid=user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_second_user(
    db: AsyncSession, mock_rgw_admin: MockRGWAdmin
) -> AsyncGenerator[UserWithAuthHeader, None]:
    """
    Create a random second user and deletes him afterwards.
    """
    user = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user.uid)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    mock_rgw_admin.delete_user(uid=user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_third_user(db: AsyncSession, mock_rgw_admin: MockRGWAdmin) -> AsyncGenerator[UserWithAuthHeader, None]:
    """
    Create a random third user and deletes him afterwards.
    """
    user = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user.uid)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    mock_rgw_admin.delete_user(uid=user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_bucket(
    db: AsyncSession, random_user: UserWithAuthHeader, mock_s3_service: MockS3ServiceResource
) -> AsyncGenerator[Bucket, None]:
    """
    Create a random user and deletes him afterwards.
    """
    bucket = await create_random_bucket(db, random_user.user)
    mock_s3_service.Bucket(name=bucket.name).create()
    mock_s3_service.BucketPolicy(bucket.name).put(
        json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "PseudoOwnerPerm",
                        "Effect": "Allow",
                        "Principal": {"AWS": [f"arn:aws:iam:::user/{random_user.user.uid}"]},
                        "Action": ["s3:GetObject", "s3:DeleteObject", "s3:PutObject", "s3:ListBucket"],
                        "Resource": [f"arn:aws:s3:::{bucket.name}/*", f"arn:aws:s3:::{bucket.name}"],
                    }
                ],
            }
        )
    )
    yield bucket
    mock_s3_service.delete_bucket(name=bucket.name, force_delete=True)
    await db.delete(bucket)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_bucket_permission(
    db: AsyncSession,
    random_second_user: UserWithAuthHeader,
    random_bucket: Bucket,
    mock_s3_service: MockS3ServiceResource,
) -> BucketPermissionDB:
    """
    Create a bucket READ permission for the second user on a bucket.
    """
    permission_db = BucketPermissionDB(user_id=random_second_user.user.uid, bucket_name=random_bucket.name)
    db.add(permission_db)
    await db.commit()
    await db.refresh(permission_db)
    mock_s3_service.Bucket(random_bucket.name).Policy().put(
        json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": BucketPermissionSchema.from_db_model(permission_db).map_to_bucket_policy_statement(
                    random_second_user.user.uid
                ),
            }
        )
    )
    return permission_db


@pytest_asyncio.fixture(scope="function")
async def random_bucket_permission_schema(
    random_bucket_permission: BucketPermissionDB, random_second_user: UserWithAuthHeader
) -> BucketPermissionSchema:
    """
    Create a bucket READ permission for the second user on a bucket.
    """

    return BucketPermissionSchema.from_db_model(random_bucket_permission, random_second_user.user.uid)
